@@ -1,19 +1,13 @@
-# -*- coding: utf-8 -*-
-
 """Define the Model class."""
 
-from __future__ import absolute_import
 
 import logging
-import types
 from copy import copy, deepcopy
 from functools import partial
 from warnings import warn
 
 import optlang
-import six
 from optlang.symbolics import Basic, Zero
-from six import iteritems, string_types
 
 from cobra.core.configuration import Configuration
 from cobra.core.dictlist import DictList
@@ -23,16 +17,15 @@ from cobra.core.metabolite import Metabolite
 from cobra.core.object import Object
 from cobra.core.reaction import Reaction
 from cobra.core.solution import get_solution
-from cobra.exceptions import SolverNotFound
 from cobra.medium import find_boundary_types, find_external_compartment, sbo_terms
 from cobra.util.context import HistoryManager, get_context, resettable
 from cobra.util.solver import (
     add_cons_vars_to_problem,
     assert_optimal,
+    check_solver,
     interface_to_str,
     remove_cons_vars_from_problem,
     set_objective,
-    solvers,
 )
 from cobra.util.util import AutoVivification, format_long_string
 
@@ -73,8 +66,7 @@ class Model(Object):
     """
 
     def __setstate__(self, state):
-        """Make sure all cobra.Objects in the model point to the model.
-        """
+        """Make sure all cobra.Objects in the model point to the model."""
         self.__dict__.update(state)
         for y in ["reactions", "genes", "metabolites"]:
             for x in getattr(self, y):
@@ -117,7 +109,7 @@ class Model(Object):
             # if not hasattr(self, '_solver'):  # backwards compatibility
             # with older cobrapy pickles?
 
-            interface = configuration.solver
+            interface = check_solver(configuration.solver)
             self._solver = interface.Model()
             self._solver.objective = interface.Objective(Zero)
             self._populate_solver(self.reactions, self.metabolites)
@@ -148,20 +140,7 @@ class Model(Object):
     @solver.setter
     @resettable
     def solver(self, value):
-        not_valid_interface = SolverNotFound(
-            "%s is not a valid solver interface. Pick from %s." % (value, list(solvers))
-        )
-        if isinstance(value, six.string_types):
-            try:
-                interface = solvers[interface_to_str(value)]
-            except KeyError:
-                raise not_valid_interface
-        elif isinstance(value, types.ModuleType) and hasattr(value, "Model"):
-            interface = value
-        elif isinstance(value, optlang.interface.Model):
-            interface = value.interface
-        else:
-            raise not_valid_interface
+        interface = check_solver(value)
 
         # Do nothing if the solver did not change
         if self.problem == interface:
@@ -174,27 +153,30 @@ class Model(Object):
 
     @tolerance.setter
     def tolerance(self, value):
-        solver_tolerances = self._solver.configuration.tolerances
+        solver_tolerances = self.solver.configuration.tolerances
 
         try:
             solver_tolerances.feasibility = value
         except AttributeError:
             logger.info(
-                "The current solver doesn't allow setting" "feasibility tolerance."
+                f"The current solver interface {interface_to_str(self.problem)} "
+                f"doesn't support setting the feasibility tolerance."
             )
 
         try:
             solver_tolerances.optimality = value
         except AttributeError:
             logger.info(
-                "The current solver doesn't allow setting" "optimality tolerance."
+                f"The current solver interface {interface_to_str(self.problem)} "
+                f"doesn't support setting the optimality tolerance."
             )
 
         try:
             solver_tolerances.integrality = value
         except AttributeError:
             logger.info(
-                "The current solver doesn't allow setting" "integrality tolerance."
+                f"The current solver interface {interface_to_str(self.problem)} "
+                f"doesn't support setting the integrality tolerance."
             )
 
         self._tolerance = value
@@ -293,10 +275,10 @@ class Model(Object):
         # Set the given media bounds
         media_rxns = list()
         exchange_rxns = frozenset(self.exchanges)
-        for rxn_id, bound in iteritems(medium):
+        for rxn_id, bound in medium.items():
             rxn = self.reactions.get_by_id(rxn_id)
             if rxn not in exchange_rxns:
-                logger.warn(
+                logger.warning(
                     "%s does not seem to be an"
                     " an exchange reaction. Applying bounds anyway.",
                     rxn.id,
@@ -308,7 +290,10 @@ class Model(Object):
 
         # Turn off reactions not present in media
         for rxn in exchange_rxns - media_rxns:
-            set_active_bound(rxn, 0)
+            is_export = rxn.reactants and not rxn.products
+            set_active_bound(
+                rxn, min(0, -rxn.lower_bound if is_export else rxn.upper_bound)
+            )
 
     def __add__(self, other_model):
         """Add the content of another model to this model (+).
@@ -355,7 +340,7 @@ class Model(Object):
         do_not_copy_by_ref = {"_reaction", "_model"}
         for metabolite in self.metabolites:
             new_met = metabolite.__class__()
-            for attr, value in iteritems(metabolite.__dict__):
+            for attr, value in metabolite.__dict__.items():
                 if attr not in do_not_copy_by_ref:
                     new_met.__dict__[attr] = copy(value) if attr == "formula" else value
             new_met._model = new
@@ -364,7 +349,7 @@ class Model(Object):
         new.genes = DictList()
         for gene in self.genes:
             new_gene = gene.__class__(None)
-            for attr, value in iteritems(gene.__dict__):
+            for attr, value in gene.__dict__.items():
                 if attr not in do_not_copy_by_ref:
                     new_gene.__dict__[attr] = (
                         copy(value) if attr == "formula" else value
@@ -376,13 +361,13 @@ class Model(Object):
         do_not_copy_by_ref = {"_model", "_metabolites", "_genes"}
         for reaction in self.reactions:
             new_reaction = reaction.__class__()
-            for attr, value in iteritems(reaction.__dict__):
+            for attr, value in reaction.__dict__.items():
                 if attr not in do_not_copy_by_ref:
                     new_reaction.__dict__[attr] = copy(value)
             new_reaction._model = new
             new.reactions.append(new_reaction)
             # update awareness
-            for metabolite, stoic in iteritems(reaction._metabolites):
+            for metabolite, stoic in reaction._metabolites.items():
                 new_met = new.metabolites.get_by_id(metabolite.id)
                 new_reaction._metabolites[new_met] = stoic
                 new_met._reaction.add(new_reaction)
@@ -397,7 +382,7 @@ class Model(Object):
         # then update their members.
         for group in self.groups:
             new_group = group.__class__(group.id)
-            for attr, value in iteritems(group.__dict__):
+            for attr, value in group.__dict__.items():
                 if attr not in do_not_copy_by_ref:
                     new_group.__dict__[attr] = copy(value)
             new_group._model = new
@@ -456,9 +441,7 @@ class Model(Object):
         metabolite_list = [x for x in metabolite_list if x.id not in self.metabolites]
 
         bad_ids = [
-            m
-            for m in metabolite_list
-            if not isinstance(m.id, string_types) or len(m.id) < 1
+            m for m in metabolite_list if not isinstance(m.id, str) or len(m.id) < 1
         ]
         if len(bad_ids) != 0:
             raise ValueError("invalid identifiers in {}".format(repr(bad_ids)))
@@ -566,7 +549,8 @@ class Model(Object):
         A demand reaction is an irreversible reaction that consumes an
         intracellular metabolite.
         A sink is similar to an exchange but specifically for intracellular
-        metabolites.
+        metabolites, i.e., a reversible reaction that adds or removes an
+        intracellular metabolite.
 
         If you set the reaction `type` to something else, you must specify the
         desired identifier of the created reaction along with its upper and
@@ -742,7 +726,7 @@ class Model(Object):
             Remove orphaned genes and metabolites from the model as well
 
         """
-        if isinstance(reactions, string_types) or hasattr(reactions, "id"):
+        if isinstance(reactions, str) or hasattr(reactions, "id"):
             warn("need to pass in a list")
             reactions = [reactions]
 
@@ -826,7 +810,7 @@ class Model(Object):
                 return False
             return True
 
-        if isinstance(group_list, string_types) or hasattr(group_list, "id"):
+        if isinstance(group_list, str) or hasattr(group_list, "id"):
             warn("need to pass in a list")
             group_list = [group_list]
 
@@ -862,7 +846,7 @@ class Model(Object):
             A list of `cobra.Group` objects to remove from the model.
         """
 
-        if isinstance(group_list, string_types) or hasattr(group_list, "id"):
+        if isinstance(group_list, str) or hasattr(group_list, "id"):
             warn("need to pass in a list")
             group_list = [group_list]
 
@@ -1027,7 +1011,7 @@ class Model(Object):
                 reaction = self.reactions.get_by_id(reaction.id)
                 forward_variable = reaction.forward_variable
                 reverse_variable = reaction.reverse_variable
-            for metabolite, coeff in six.iteritems(reaction.metabolites):
+            for metabolite, coeff in reaction.metabolites.items():
                 if metabolite.id in self.constraints:
                     constraint = self.constraints[metabolite.id]
                 else:
@@ -1042,7 +1026,7 @@ class Model(Object):
         for reaction in reaction_list:
             reaction = self.reactions.get_by_id(reaction.id)
             reaction.update_variable_bounds()
-        for constraint, terms in six.iteritems(constraint_terms):
+        for constraint, terms in constraint_terms.items():
             constraint.set_linear_coefficients(terms)
 
     def slim_optimize(self, error_value=float("nan"), message=None):
@@ -1133,6 +1117,7 @@ class Model(Object):
             for gene in self.genes:
                 gene._reaction.clear()
             for rxn in self.reactions:
+                rxn._update_genes_from_gpr()
                 for met in rxn._metabolites:
                     met._reaction.add(rxn)
                 for gene in rxn._genes:
@@ -1332,7 +1317,5 @@ class Model(Object):
             num_reactions=len(self.reactions),
             num_groups=len(self.groups),
             objective=format_long_string(str(self.objective.expression), 100),
-            compartments=", ".join(
-                v if v else k for k, v in iteritems(self.compartments)
-            ),
+            compartments=", ".join(v if v else k for k, v in self.compartments.items()),
         )
