@@ -11,13 +11,13 @@ from copy import copy, deepcopy
 from functools import partial
 from math import isinf
 from operator import attrgetter
+from typing import FrozenSet
 from warnings import warn
 
 from future.utils import raise_from, raise_with_traceback
-from six import iteritems, iterkeys, string_types
 
 from cobra.core.configuration import Configuration
-from cobra.core.gene import Gene, ast2str, eval_gpr, parse_gpr
+from cobra.core.gene import GPR, Gene
 from cobra.core.metabolite import Metabolite
 from cobra.core.object import Object
 from cobra.exceptions import OptimizationError
@@ -40,7 +40,7 @@ uppercase_OR = re.compile(r"\bOR\b")
 gpr_clean = re.compile(" {2,}")
 # This regular expression finds any single letter compartment enclosed in
 # square brackets at the beginning of the string. For example [c] : foo --> bar
-compartment_finder = re.compile("^\s*(\[[A-Za-z]\])\s*:*")
+compartment_finder = re.compile(r"^\s*(\[[A-Za-z]\])\s*:*")
 # Regular expressions to match the arrows
 _reversible_arrow_finder = re.compile("<(-+|=+)>")
 _forward_arrow_finder = re.compile("(-+|=+)>")
@@ -76,7 +76,7 @@ class Reaction(Object):
         self, id=None, name="", subsystem="", lower_bound=0.0, upper_bound=None
     ):
         Object.__init__(self, id, name)
-        self._gene_reaction_rule = ""
+        self._gpr = GPR()
         self.subsystem = subsystem
 
         # The cobra.Genes that are used to catalyze the reaction
@@ -86,9 +86,6 @@ class Reaction(Object):
         # this reaction.
         self._metabolites = {}
 
-        # The set of compartments that partaking metabolites are in.
-        self._compartments = None
-
         # self.model is None or refers to the cobra.Model that
         # contains self
         self._model = None
@@ -130,7 +127,7 @@ class Reaction(Object):
             The expression representing the the forward flux (if associated
             with model), otherwise None. Representing the net flux if
             model.reversible_encoding == 'unsplit' or None if reaction is
-            not associated with a model """
+            not associated with a model"""
         if self.model is not None:
             return 1.0 * self.forward_variable - 1.0 * self.reverse_variable
         else:
@@ -169,7 +166,7 @@ class Reaction(Object):
 
     @property
     def objective_coefficient(self):
-        """ Get the coefficient for this reaction in a linear
+        """Get the coefficient for this reaction in a linear
         objective (float)
 
         Assuming that the objective of the associated model is summation of
@@ -243,17 +240,6 @@ class Reaction(Object):
     @lower_bound.setter
     @resettable
     def lower_bound(self, value):
-        if self._upper_bound < value:
-            warn(
-                "You are constraining the reaction '{}' to a fixed flux "
-                "value of {}. Did you intend to do this? We are planning to "
-                "remove this behavior in a future release. Please let us "
-                "know your opinion at "
-                "https://github.com/opencobra/cobrapy/issues/793."
-                "".format(self.id, value),
-                DeprecationWarning,
-            )
-            self.upper_bound = value
         # Validate bounds before setting them.
         self._check_bounds(value, self._upper_bound)
         self._lower_bound = value
@@ -276,17 +262,6 @@ class Reaction(Object):
     @upper_bound.setter
     @resettable
     def upper_bound(self, value):
-        if self._lower_bound > value:
-            warn(
-                "You are constraining the reaction '{}' to a fixed flux "
-                "value of {}. Did you intend to do this? We are planning to "
-                "remove this behavior in a future release. Please let us "
-                "know your opinion at "
-                "https://github.com/opencobra/cobrapy/issues/793."
-                "".format(self.id, value),
-                DeprecationWarning,
-            )
-            self.lower_bound = value
         # Validate bounds before setting them.
         self._check_bounds(self._lower_bound, value)
         self._upper_bound = value
@@ -294,7 +269,7 @@ class Reaction(Object):
 
     @property
     def bounds(self):
-        """ Get or set the bounds directly from a tuple
+        """Get or set the bounds directly from a tuple
 
         Convenience method for setting upper and lower bounds in one line
         using a tuple of lower and upper bound. Invalid bounds will raise an
@@ -434,49 +409,38 @@ class Reaction(Object):
         return self._metabolites.copy()
 
     @property
-    def genes(self):
+    def genes(self) -> FrozenSet:
+        """Return the genes of the reaction.
+
+        Returns
+        -------
+        genes: FrozenSet
+        """
         return frozenset(self._genes)
 
-    @property
-    def gene_reaction_rule(self):
-        return self._gene_reaction_rule
+    def _update_genes_from_gpr(self, new_gene_names: set = None) -> None:
+        """Update genes of reation based on GPR.
 
-    @gene_reaction_rule.setter
-    def gene_reaction_rule(self, new_rule):
-
-        # TODO: Do this :)
-        if get_context(self):
-            warn("Context management not implemented for " "gene reaction rules")
-
-        self._gene_reaction_rule = new_rule.strip()
-        try:
-            _, gene_names = parse_gpr(self._gene_reaction_rule)
-        except (SyntaxError, TypeError) as e:
-            if "AND" in new_rule or "OR" in new_rule:
-                warn(
-                    "uppercase AND/OR found in rule '%s' for '%s'"
-                    % (new_rule, repr(self))
-                )
-                new_rule = uppercase_AND.sub("and", new_rule)
-                new_rule = uppercase_OR.sub("or", new_rule)
-                self.gene_reaction_rule = new_rule
-                return
-            warn("malformed gene_reaction_rule '%s' for %s" % (new_rule, repr(self)))
-            tmp_str = and_or_search.sub("", self._gene_reaction_rule)
-            gene_names = set((gpr_clean.sub(" ", tmp_str).split(" ")))
-        if "" in gene_names:
-            gene_names.remove("")
-        old_genes = self._genes
+        Parameters
+        ----------
+        new_gene_names: set
+        """
+        if new_gene_names is None:
+            if self._gpr is not None:
+                new_gene_names = self._gpr.genes
+            else:
+                new_gene_names = set()
+        old_genes = self._genes.copy()
         if self._model is None:
-            self._genes = {Gene(i) for i in gene_names}
+            self._genes = {Gene(i) for i in new_gene_names}
         else:
             model_genes = self._model.genes
             self._genes = set()
-            for id in gene_names:
-                if model_genes.has_id(id):
-                    self._genes.add(model_genes.get_by_id(id))
+            for g_id in new_gene_names:
+                if model_genes.has_id(g_id):
+                    self._genes.add(model_genes.get_by_id(g_id))
                 else:
-                    new_gene = Gene(id)
+                    new_gene = Gene(g_id)
                     new_gene._model = self._model
                     self._genes.add(new_gene)
                     model_genes.append(new_gene)
@@ -496,6 +460,37 @@ class Reaction(Object):
                         % (g.id, self.id)
                     )
 
+    @property
+    def gene_reaction_rule(self) -> str:
+        """See gene reaction rule as string.
+
+        Uses the to_string() method of the GPR class
+
+        Returns
+        -------
+        string
+
+        """
+        return self._gpr.to_string()
+
+    @gene_reaction_rule.setter
+    def gene_reaction_rule(self, new_rule: str):
+        """Set a new GPR for the reaction, using a string expression.
+
+        Parameters
+        ----------
+        new_rule : string
+            which will be parsed by the string parser in GPR, GPR.from_string(new_rule).
+            It makes a new GPR, and does not modify the existing one.
+
+        """
+        # TODO: Do this :)
+        if get_context(self):
+            warn("Context management not implemented for gene reaction rules.")
+
+        self._gpr = GPR.from_string(new_rule)
+        self._update_genes_from_gpr()
+
     @property
     def gene_name_reaction_rule(self):
         """Display gene_reaction_rule with names intead.
@@ -506,8 +501,24 @@ class Reaction(Object):
 
         """
         names = {i.id: i.name for i in self._genes}
-        ast = parse_gpr(self._gene_reaction_rule)[0]
-        return ast2str(ast, names=names)
+        return self._gpr.to_string(names=names)
+
+    @property
+    def gpr(self):
+        """Return the GPR associated with the reaction."""
+        return self._gpr
+
+    @gpr.setter
+    def gpr(self, value: GPR):
+        """Set a new GPR for the reaction, using GPR() class.
+
+        Parameters
+        ----------
+        value : GPR() class.
+
+        """
+        self._gpr = value
+        self._update_genes_from_gpr()
 
     @property
     def functional(self):
@@ -521,9 +532,8 @@ class Reaction(Object):
             otherwise False.
         """
         if self._model:
-            tree, _ = parse_gpr(self.gene_reaction_rule)
-            return eval_gpr(
-                tree, {gene.id for gene in self.genes if not gene.functional}
+            return self._gpr.eval(
+                {gene.id for gene in self.genes if not gene.functional}
             )
         return True
 
@@ -725,7 +735,7 @@ class Reaction(Object):
         bounds are swapped.
         """
         self._metabolites = {
-            met: value * coefficient for met, value in iteritems(self._metabolites)
+            met: value * coefficient for met, value in self._metabolites.items()
         }
 
         if coefficient < 0:
@@ -749,12 +759,12 @@ class Reaction(Object):
     @property
     def reactants(self):
         """Return a list of reactants for the reaction."""
-        return [k for k, v in iteritems(self._metabolites) if v < 0]
+        return [k for k, v in self._metabolites.items() if v < 0]
 
     @property
     def products(self):
         """Return a list of products for the reaction"""
-        return [k for k, v in iteritems(self._metabolites) if v >= 0]
+        return [k for k, v in self._metabolites.items() if v >= 0]
 
     def get_coefficient(self, metabolite_id):
         """
@@ -812,7 +822,7 @@ class Reaction(Object):
         new_metabolites = []
         _id_to_metabolites = dict([(x.id, x) for x in self._metabolites])
 
-        for metabolite, coefficient in iteritems(metabolites_to_add):
+        for metabolite, coefficient in metabolites_to_add.items():
 
             # Make sure metabolites being added belong to the same model, or
             # else copy them.
@@ -843,7 +853,7 @@ class Reaction(Object):
                         else:
                             # do we want to handle creation here?
                             raise e
-                elif isinstance(metabolite, string_types):
+                elif isinstance(metabolite, str):
                     # if we want to handle creation, this should be changed
                     raise ValueError(
                         "Reaction '%s' does not belong to a "
@@ -892,7 +902,7 @@ class Reaction(Object):
                 # Reset them with add_metabolites
                 mets_to_reset = {
                     key: old_coefficients[model.metabolites.get_by_any(key)[0]]
-                    for key in iterkeys(metabolites_to_add)
+                    for key in metabolites_to_add.keys()
                 }
 
                 context(
@@ -934,7 +944,7 @@ class Reaction(Object):
 
         """
         self.add_metabolites(
-            {k: -v for k, v in iteritems(metabolites)},
+            {k: -v for k, v in metabolites.items()},
             combine=combine,
             reversibly=reversibly,
         )
@@ -986,26 +996,22 @@ class Reaction(Object):
         This should be empty for balanced reactions.
         """
         reaction_element_dict = defaultdict(int)
-        for metabolite, coefficient in iteritems(self._metabolites):
+        for metabolite, coefficient in self._metabolites.items():
             if metabolite.charge is not None:
                 reaction_element_dict["charge"] += coefficient * metabolite.charge
             if metabolite.elements is None:
                 raise ValueError("No elements found in metabolite %s" % metabolite.id)
-            for element, amount in iteritems(metabolite.elements):
+            for element, amount in metabolite.elements.items():
                 reaction_element_dict[element] += coefficient * amount
         # filter out 0 values
-        return {k: v for k, v in iteritems(reaction_element_dict) if v != 0}
+        return {k: v for k, v in reaction_element_dict.items() if v != 0}
 
     @property
     def compartments(self):
         """lists compartments the metabolites are in"""
-        if self._compartments is None:
-            self._compartments = {
-                met.compartment
-                for met in self._metabolites
-                if met.compartment is not None
-            }
-        return self._compartments
+        return {
+            met.compartment for met in self._metabolites if met.compartment is not None
+        }
 
     def get_compartments(self):
         """lists compartments the metabolites are in"""
@@ -1174,7 +1180,10 @@ class Reaction(Object):
         from cobra.summary import ReactionSummary
 
         return ReactionSummary(
-            reaction=self, model=self._model, solution=solution, fva=fva,
+            reaction=self,
+            model=self._model,
+            solution=solution,
+            fva=fva,
         )
 
     def __str__(self):
