@@ -33,29 +33,22 @@ import datetime
 import logging
 import os
 import re
-import traceback
+from ast import And, BoolOp, Module, Name, Or
 from collections import defaultdict, namedtuple
 from copy import deepcopy
+from io import StringIO
 from sys import platform
 
 import libsbml
-from six import iteritems, raise_from, string_types
 
 import cobra
-from cobra.core import Gene, Group, Metabolite, Model, Reaction
-from cobra.core.gene import parse_gpr
+from cobra.core import GPR, Gene, Group, Metabolite, Model, Reaction
 from cobra.manipulation.validate import check_metabolite_compartment_formula
 from cobra.util.solver import linear_reaction_coefficients, set_objective
 
 
-try:
-    from cStringIO import StringIO  # Python 2
-except ImportError:
-    from io import StringIO
-
-
 class CobraSBMLError(Exception):
-    """ SBML error class. """
+    """SBML error class."""
 
     pass
 
@@ -111,12 +104,12 @@ pattern_from_sbml = re.compile(r"__(\d+)__")
 
 def _escape_non_alphanum(nonASCII):
     """converts a non alphanumeric character to a string representation of
-    its ascii number """
+    its ascii number"""
     return "__" + str(ord(nonASCII.group())) + "__"
 
 
 def _number_to_chr(numberStr):
-    """converts an ascii number to a character """
+    """converts an ascii number to a character"""
     return chr(int(numberStr.group(1)))
 
 
@@ -262,7 +255,7 @@ def read_sbml_model(filename, number=float, f_replace=F_REPLACE, **kwargs):
             "\nIf the model is valid and cannot be read please open an issue "
             "at https://github.com/opencobra/cobrapy/issues ."
         )
-        raise_from(cobra_error, original_error)
+        raise cobra_error from original_error
 
 
 def _get_doc_from_filename(filename):
@@ -276,7 +269,7 @@ def _get_doc_from_filename(filename):
     -------
     libsbml.SBMLDocument
     """
-    if isinstance(filename, string_types):
+    if isinstance(filename, str):
         if ("win" in platform) and (len(filename) < 260) and os.path.exists(filename):
             # path (win)
             doc = libsbml.readSBMLFromFile(
@@ -567,35 +560,24 @@ def _sbml_to_model(
 
     # GPR rules
     def process_association(ass):
-        """ Recursively convert gpr association to a gpr string.
+        """Recursively convert gpr association to a GPR class.
+
         Defined as inline functions to not pass the replacement dict around.
         """
         if ass.isFbcOr():
-            return " ".join(
-                [
-                    "(",
-                    " or ".join(
-                        process_association(c) for c in ass.getListOfAssociations()
-                    ),
-                    ")",
-                ]
+            return BoolOp(
+                Or(), [process_association(c) for c in ass.getListOfAssociations()]
             )
         elif ass.isFbcAnd():
-            return " ".join(
-                [
-                    "(",
-                    " and ".join(
-                        process_association(c) for c in ass.getListOfAssociations()
-                    ),
-                    ")",
-                ]
+            return BoolOp(
+                And(), [process_association(c) for c in ass.getListOfAssociations()]
             )
         elif ass.isGeneProductRef():
             gid = ass.getGeneProduct()
             if f_replace and F_GENE in f_replace:
-                return f_replace[F_GENE](gid)
+                return Name(id=f_replace[F_GENE](gid))
             else:
-                return gid
+                return Name(id=gid)
 
     # Reactions
     missing_bounds = False
@@ -723,7 +705,10 @@ def _sbml_to_model(
                 association = (
                     gpa.getAssociation()
                 )  # noqa: E501 type: libsbml.FbcAssociation
-                gpr = process_association(association)
+                gpr = Module(process_association(association))
+            else:
+                gpr = None
+            cobra_reaction.gpr = GPR(gpr_from=gpr)
         else:
             # fallback to notes information
             notes = cobra_reaction.notes
@@ -743,20 +728,7 @@ def _sbml_to_model(
                 )
                 if f_replace and F_GENE in f_replace:
                     gpr = " ".join(f_replace[F_GENE](t) for t in gpr.split(" "))
-
-        # remove outside parenthesis, if any
-        if gpr.startswith("(") and gpr.endswith(")"):
-            try:
-                parse_gpr(gpr[1:-1].strip())
-                gpr = gpr[1:-1].strip()
-            except (SyntaxError, TypeError) as e:
-                LOGGER.warning(
-                    "Removing parenthesis from gpr %s leads to "
-                    "an error, so keeping parenthesis",
-                    gpr,
-                )
-
-        cobra_reaction.gene_reaction_rule = gpr
+            cobra_reaction.gpr = GPR.from_string(gpr)
 
     cobra_model.add_reactions(reactions)
 
@@ -963,7 +935,7 @@ def write_sbml_model(cobra_model, filename, f_replace=F_REPLACE, **kwargs):
     """
     doc = _model_to_sbml(cobra_model, f_replace=f_replace, **kwargs)
 
-    if isinstance(filename, string_types):
+    if isinstance(filename, str):
         # write to path
         libsbml.writeSBMLToFile(doc, filename)
 
@@ -1090,7 +1062,7 @@ def _model_to_sbml(cobra_model, f_replace=None, units=True):
     # Compartments
     # FIXME: use first class compartment model (and write notes & annotations)
     #     (https://github.com/opencobra/cobrapy/issues/811)
-    for cid, name in iteritems(cobra_model.compartments):
+    for cid, name in cobra_model.compartments.items():
         compartment = model.createCompartment()  # type: libsbml.Compartment
         compartment.setId(cid)
         compartment.setName(name)
@@ -1158,7 +1130,7 @@ def _model_to_sbml(cobra_model, f_replace=None, units=True):
         _sbase_notes_dict(reaction, cobra_reaction.notes)
 
         # stoichiometry
-        for metabolite, stoichiometry in iteritems(cobra_reaction._metabolites):
+        for metabolite, stoichiometry in cobra_reaction._metabolites.items():
             sid = metabolite.id
             if f_replace and F_SPECIE_REV in f_replace:
                 sid = f_replace[F_SPECIE_REV](sid)
@@ -1201,21 +1173,14 @@ def _model_to_sbml(cobra_model, f_replace=None, units=True):
         )
 
         # GPR
-        gpr = cobra_reaction.gene_reaction_rule
-        if gpr is not None and len(gpr) > 0:
-
+        gpr = cobra_reaction.gpr
+        if gpr is not None and gpr.body:
             # replace ids in string
             if f_replace and F_GENE_REV in f_replace:
-                gpr = gpr.replace("(", "( ")
-                gpr = gpr.replace(")", " )")
-                tokens = gpr.split()
-
-                for k in range(len(tokens)):
-                    if tokens[k] not in ["and", "or", "(", ")"]:
-                        tokens[k] = f_replace[F_GENE_REV](tokens[k])
-                gpr_new = " ".join(tokens)
+                idmap = {gid: f_replace[F_GENE_REV](gid) for gid in gpr.genes}
+                gpr_new = gpr.to_string(names=idmap)
             else:
-                gpr_new = gpr
+                gpr_new = gpr.to_string()
 
             gpa = (
                 r_fbc.createGeneProductAssociation()
@@ -1406,7 +1371,7 @@ def _check(value, message):
 # Notes
 # -----------------------------------------------------------------------------
 def _parse_notes_dict(sbase):
-    """ Creates dictionary of COBRA notes.
+    """Creates dictionary of COBRA notes.
 
     Parameters
     ----------
@@ -1544,7 +1509,7 @@ def _parse_annotations(sbase):
                 provider, identifier = data
 
             if provider in annotation:
-                if isinstance(annotation[provider], string_types):
+                if isinstance(annotation[provider], str):
                     annotation[provider] = [annotation[provider]]
                 # FIXME: use a list
                 if identifier not in annotation[provider]:
@@ -1610,12 +1575,12 @@ def _sbase_annotations(sbase, annotation):
         # handling of non-string annotations (e.g. integers)
         if isinstance(value, (float, int)):
             value = str(value)
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             annotation_data[key] = [("is", value)]
 
     for key, value in annotation_data.items():
         for idx, item in enumerate(value):
-            if isinstance(item, string_types):
+            if isinstance(item, str):
                 value[idx] = ("is", item)
 
     # set metaId
@@ -1623,7 +1588,7 @@ def _sbase_annotations(sbase, annotation):
     sbase.setMetaId(meta_id)
 
     # rdf_items = []
-    for provider, data in iteritems(annotation_data):
+    for provider, data in annotation_data.items():
 
         # set SBOTerm
         if provider in ["SBO", "sbo"]:
